@@ -30,9 +30,11 @@
 
 namespace OrthancPlugins
 {
-  PostgreSQLWrapper::PostgreSQLWrapper(PostgreSQLConnection* connection,
+  PostgreSQLWrapper::PostgreSQLWrapper(OrthancPluginContext* context,
+                                       PostgreSQLConnection* connection,
                                        bool useLock,
                                        bool allowUnlock) :
+    context_(context),
     connection_(connection),
     globalProperties_(*connection, useLock, GlobalProperty_IndexLock)
   {
@@ -108,38 +110,65 @@ namespace OrthancPlugins
 
   void PostgreSQLWrapper::Prepare()
   {
+    uint32_t expectedVersion;
+    if (context_)
+    {
+      expectedVersion = OrthancPluginGetExpectedDatabaseVersion(context_);
+    }
+    else
+    {
+      // This case only occurs during unit testing
+      expectedVersion = 6;
+    }
+
+    /* Check the expected version of the database */
+    if (expectedVersion != 5 && expectedVersion != 6)
+    {
+      char info[1024];
+      sprintf(info, "This database plugin is incompatible with your version of Orthanc "
+              "expecting the DB schema version %d, but this plugin is compatible with versions 5 or 6",
+              expectedVersion);
+      OrthancPluginLogError(context_, info);
+      throw PostgreSQLException(info);
+    }
+
+
     PostgreSQLTransaction t(*connection_);
 
     if (!connection_->DoesTableExist("Resources"))
     {
       std::string query;
-      EmbeddedResources::GetFileResource(query, EmbeddedResources::POSTGRESQL_PREPARE);
+
+      if (expectedVersion == 5)
+      {
+        EmbeddedResources::GetFileResource(query, EmbeddedResources::POSTGRESQL_PREPARE_V5);
+      }
+      else
+      {
+        EmbeddedResources::GetFileResource(query, EmbeddedResources::POSTGRESQL_PREPARE_V6);
+      }
 
       connection_->Execute(query);
-    }
 
+      // Execute the common initialization between versions 5 and 6
+      EmbeddedResources::GetFileResource(query, EmbeddedResources::POSTGRESQL_PREPARE);
+      connection_->Execute(query);
+
+      // Set the version of the database schema
+      // The "1" corresponds to the "GlobalProperty_DatabaseSchemaVersion" enumeration
+      connection_->Execute("INSERT INTO GlobalProperties VALUES (1, '" + 
+                           boost::lexical_cast<std::string>(expectedVersion) + "')");
+
+    }
     
     // Check the version of the database
-    std::string version = "unknown";
-    if (!LookupGlobalProperty(version, GlobalProperty_DatabaseSchemaVersion))
-    {
-      throw PostgreSQLException("The database is corrupted. Drop it manually for Orthanc to recreate it");
-    }
+    version_ = GetDatabaseVersion();
 
-    bool ok = false;
-
-    try
+    if (version_ != 5 &&
+        version_ != 6)
     {
-      unsigned int v = boost::lexical_cast<unsigned int>(version);
-      ok = (v == 5);
-    }
-    catch (boost::bad_lexical_cast&)
-    {
-    }
-   
-    if (!ok)
-    {
-      std::string message = "Incompatible version of the Orthanc PostgreSQL database: " + version;
+      std::string message = ("Incompatible version of the Orthanc PostgreSQL database: " + 
+                             boost::lexical_cast<std::string>(version_));
       throw PostgreSQLException(message);
     }
           
@@ -340,6 +369,61 @@ namespace OrthancPlugins
   }
 
 
+  void PostgreSQLWrapper::GetAllInternalIds(std::list<int64_t>& target,
+                                            OrthancPluginResourceType resourceType)
+  {
+    if (getAllInternalIds_.get() == NULL)
+    {
+      getAllInternalIds_.reset
+        (new PostgreSQLStatement(*connection_, 
+                                 "SELECT internalId FROM Resources WHERE resourceType=$1"));
+      getAllInternalIds_->DeclareInputInteger(0);
+    }
+
+    getAllInternalIds_->BindInteger(0, static_cast<int>(resourceType));
+    PostgreSQLResult result(*getAllInternalIds_);
+
+    target.clear();
+
+    while (!result.IsDone())
+    {
+      target.push_back(result.GetInteger64(0));
+      result.Step();
+    }
+  }
+
+
+  void PostgreSQLWrapper::GetAllPublicIds(std::list<std::string>& target,
+                                          OrthancPluginResourceType resourceType,
+                                          uint64_t since,
+                                          uint64_t limit)
+  {
+    if (getAllPublicIdsWithLimit_.get() == NULL)
+    {
+      getAllPublicIdsWithLimit_.reset
+        (new PostgreSQLStatement(*connection_, 
+                                 "SELECT * FROM (SELECT publicId FROM Resources WHERE resourceType=$1) AS tmp ORDER BY tmp.publicId LIMIT $2 OFFSET $3"));
+
+      getAllPublicIdsWithLimit_->DeclareInputInteger(0);
+      getAllPublicIdsWithLimit_->DeclareInputInteger64(1);
+      getAllPublicIdsWithLimit_->DeclareInputInteger64(2);
+    }
+
+    getAllPublicIdsWithLimit_->BindInteger(0, static_cast<int>(resourceType));
+    getAllPublicIdsWithLimit_->BindInteger64(1, limit);
+    getAllPublicIdsWithLimit_->BindInteger64(2, since);
+
+    PostgreSQLResult result(*getAllPublicIdsWithLimit_);
+    target.clear();
+
+    while (!result.IsDone())
+    {
+      target.push_back(result.GetString(0));
+      result.Step();
+    }
+  }
+
+
   void PostgreSQLWrapper::GetChangesInternal(bool& done,
                                              PostgreSQLStatement& s,
                                              uint32_t maxResults)
@@ -536,6 +620,7 @@ namespace OrthancPlugins
       }
     }
 
+    if (version_ == 5)
     {
       getMainDicomTags2_->BindInteger64(0, id);
       PostgreSQLResult result(*getMainDicomTags2_);
@@ -839,6 +924,7 @@ namespace OrthancPlugins
   }
 
 
+  // Used only if Orthanc <= 0.9.4
   void PostgreSQLWrapper::LookupIdentifier(std::list<int64_t>& target,
                                            uint16_t group,
                                            uint16_t element,
@@ -851,7 +937,15 @@ namespace OrthancPlugins
          (*connection_, "SELECT id FROM DicomIdentifiers WHERE tagGroup=$1 AND tagElement=$2 and value=$3"));
       lookupIdentifier1_->DeclareInputInteger(0);
       lookupIdentifier1_->DeclareInputInteger(1);
-      lookupIdentifier1_->DeclareInputBinary(2);
+
+      if (version_ == 5)
+      {
+        lookupIdentifier1_->DeclareInputBinary(2);
+      }
+      else
+      {
+        lookupIdentifier1_->DeclareInputString(2);
+      }
     }
 
 
@@ -869,6 +963,8 @@ namespace OrthancPlugins
     }
   }
 
+
+  // Used only if Orthanc <= 0.9.4
   void PostgreSQLWrapper::LookupIdentifier(std::list<int64_t>& target,
                                            const char* value)
   {
@@ -877,7 +973,15 @@ namespace OrthancPlugins
       lookupIdentifier2_.reset
         (new PostgreSQLStatement
          (*connection_, "SELECT id FROM DicomIdentifiers WHERE value=$1"));
-      lookupIdentifier2_->DeclareInputBinary(0);
+
+      if (version_ == 5)
+      {
+        lookupIdentifier2_->DeclareInputBinary(0);
+      }
+      else
+      {
+        lookupIdentifier2_->DeclareInputString(0);
+      }
     }
 
     lookupIdentifier2_->BindString(0, value);
@@ -893,6 +997,137 @@ namespace OrthancPlugins
   }
 
 
+#if !(ORTHANC_PLUGINS_MINIMAL_MAJOR_NUMBER <= 0 && ORTHANC_PLUGINS_MINIMAL_MINOR_NUMBER <= 9 && ORTHANC_PLUGINS_MINIMAL_REVISION_NUMBER <= 4)
+  static std::string ConvertWildcardToLike(const std::string& query)
+  {
+    std::string s = query;
+
+    for (size_t i = 0; i < s.size(); i++)
+    {
+      if (s[i] == '*')
+      {
+        s[i] = '%';
+      }
+      else if (s[i] == '?')
+      {
+        s[i] = '_';
+      }
+    }
+
+    return s;
+  }
+
+
+  // Used only if Orthanc >= 0.9.5
+  void PostgreSQLWrapper::LookupIdentifier(std::list<int64_t>& target,
+                                           OrthancPluginResourceType level,
+                                           uint16_t group,
+                                           uint16_t element,
+                                           OrthancPluginIdentifierConstraint constraint,
+                                           const char* value)
+  {
+    if (lookupIdentifierEQ_.get() == NULL)
+    {
+      lookupIdentifierEQ_.reset
+        (new PostgreSQLStatement
+         (*connection_, 
+          "SELECT d.id FROM DicomIdentifiers AS d, Resources AS r WHERE "
+          "d.id = r.internalId AND r.resourceType=$1 AND d.tagGroup=$2 AND d.tagElement=$3 AND d.value=$4"));
+      lookupIdentifierEQ_->DeclareInputInteger(0);
+      lookupIdentifierEQ_->DeclareInputInteger(1);
+      lookupIdentifierEQ_->DeclareInputInteger(2);
+      lookupIdentifierEQ_->DeclareInputString(3);
+    }
+
+    if (lookupIdentifierLE_.get() == NULL)
+    {
+      lookupIdentifierLE_.reset
+        (new PostgreSQLStatement
+         (*connection_, 
+          "SELECT d.id FROM DicomIdentifiers AS d, Resources AS r WHERE "
+          "d.id = r.internalId AND r.resourceType=$1 AND d.tagGroup=$2 AND d.tagElement=$3 AND d.value<=$4"));
+      lookupIdentifierLE_->DeclareInputInteger(0);
+      lookupIdentifierLE_->DeclareInputInteger(1);
+      lookupIdentifierLE_->DeclareInputInteger(2);
+      lookupIdentifierLE_->DeclareInputString(3);
+    }
+
+    if (lookupIdentifierGE_.get() == NULL)
+    {
+      lookupIdentifierGE_.reset
+        (new PostgreSQLStatement
+         (*connection_, 
+          "SELECT d.id FROM DicomIdentifiers AS d, Resources AS r WHERE "
+          "d.id = r.internalId AND r.resourceType=$1 AND d.tagGroup=$2 AND d.tagElement=$3 AND d.value>=$4"));
+      lookupIdentifierGE_->DeclareInputInteger(0);
+      lookupIdentifierGE_->DeclareInputInteger(1);
+      lookupIdentifierGE_->DeclareInputInteger(2);
+      lookupIdentifierGE_->DeclareInputString(3);
+    }
+
+    if (lookupIdentifierWildcard_.get() == NULL)
+    {
+      lookupIdentifierWildcard_.reset
+        (new PostgreSQLStatement
+         (*connection_, 
+          "SELECT d.id FROM DicomIdentifiers AS d, Resources AS r WHERE "
+          "d.id = r.internalId AND r.resourceType=$1 AND d.tagGroup=$2 AND d.tagElement=$3 AND d.value LIKE $4"));
+      lookupIdentifierWildcard_->DeclareInputInteger(0);
+      lookupIdentifierWildcard_->DeclareInputInteger(1);
+      lookupIdentifierWildcard_->DeclareInputInteger(2);
+      lookupIdentifierWildcard_->DeclareInputString(3);
+    }
+
+    PostgreSQLStatement* statement = NULL;
+
+    switch (constraint)
+    {
+      case OrthancPluginIdentifierConstraint_Equal:
+        statement = lookupIdentifierEQ_.get();
+        break;
+
+      case OrthancPluginIdentifierConstraint_SmallerOrEqual:
+        statement = lookupIdentifierLE_.get();
+        break;
+
+      case OrthancPluginIdentifierConstraint_GreaterOrEqual:
+        statement = lookupIdentifierGE_.get();
+        break;
+
+      case OrthancPluginIdentifierConstraint_Wildcard:
+        statement = lookupIdentifierWildcard_.get();
+        break;
+
+      default:
+        throw PostgreSQLException();
+    }
+
+    assert(statement != NULL);
+    statement->BindInteger(0, level);
+    statement->BindInteger(1, group);
+    statement->BindInteger(2, element);
+    if (constraint == OrthancPluginIdentifierConstraint_Wildcard)
+    {
+      statement->BindString(3, ConvertWildcardToLike(value));
+    }
+    else
+    {
+      statement->BindString(3, value);
+    }
+
+    PostgreSQLResult result(*statement);
+    target.clear();
+
+    while (!result.IsDone())
+    {
+      target.push_back(result.GetInteger64(0));
+      result.Step();
+    }
+  }
+#endif
+
+
+
   bool PostgreSQLWrapper::LookupMetadata(std::string& target,
                                          int64_t id,
                                          int32_t type)
@@ -1059,7 +1294,15 @@ namespace OrthancPlugins
       setMainDicomTags_->DeclareInputInteger64(0);
       setMainDicomTags_->DeclareInputInteger(1);
       setMainDicomTags_->DeclareInputInteger(2);
-      setMainDicomTags_->DeclareInputBinary(3);
+
+      if (version_ == 5)
+      {
+        setMainDicomTags_->DeclareInputBinary(3);
+      }
+      else
+      {
+        setMainDicomTags_->DeclareInputString(3);
+      }
     }
 
     SetTagInternal(*setMainDicomTags_, id, group, element, value);
@@ -1078,7 +1321,15 @@ namespace OrthancPlugins
       setIdentifierTag_->DeclareInputInteger64(0);
       setIdentifierTag_->DeclareInputInteger(1);
       setIdentifierTag_->DeclareInputInteger(2);
-      setIdentifierTag_->DeclareInputBinary(3);
+
+      if (version_ == 5)
+      {
+        setIdentifierTag_->DeclareInputBinary(3);
+      }
+      else
+      {
+        setIdentifierTag_->DeclareInputString(3);
+      }
     }
 
     SetTagInternal(*setIdentifierTag_, id, group, element, value);
@@ -1222,4 +1473,80 @@ namespace OrthancPlugins
       return true;
     }
   }
+
+
+  uint32_t PostgreSQLWrapper::GetDatabaseVersion()
+  {
+    // Check the version of the database
+    std::string version = "unknown";
+    if (!LookupGlobalProperty(version, GlobalProperty_DatabaseSchemaVersion))
+    {
+      throw PostgreSQLException("The database is corrupted. Drop it manually for Orthanc to recreate it");
+    }
+
+    try
+    {
+      return boost::lexical_cast<unsigned int>(version);
+    }
+    catch (boost::bad_lexical_cast&)
+    {
+      throw PostgreSQLException("The database is corrupted. Drop it manually for Orthanc to recreate it");
+    }
+  }
+
+
+  void PostgreSQLWrapper::UpgradeDatabase(uint32_t  targetVersion,
+                                          OrthancPluginStorageArea* storageArea)
+  {
+    unsigned int version = GetDatabaseVersion();
+
+#if !(ORTHANC_PLUGINS_MINIMAL_MAJOR_NUMBER <= 0 && ORTHANC_PLUGINS_MINIMAL_MINOR_NUMBER <= 9 && ORTHANC_PLUGINS_MINIMAL_REVISION_NUMBER <= 4)
+
+    if (version == 5 && targetVersion == 6)
+    {
+      version_ = 6;  // Must be here, otherwise calls to
+                     // SetMainDicomTag() during the reconstruction of
+                     // the DICOM tags will fail as the column type
+                     // has been altered
+      connection_->Execute("ALTER TABLE MainDicomTags ALTER COLUMN value TYPE TEXT");
+      connection_->Execute("ALTER TABLE DicomIdentifiers ALTER COLUMN value TYPE TEXT");
+      OrthancPluginReconstructMainDicomTags(context_, storageArea, OrthancPluginResourceType_Patient);
+      OrthancPluginReconstructMainDicomTags(context_, storageArea, OrthancPluginResourceType_Study);
+      OrthancPluginReconstructMainDicomTags(context_, storageArea, OrthancPluginResourceType_Series);
+      OrthancPluginReconstructMainDicomTags(context_, storageArea, OrthancPluginResourceType_Instance);
+      connection_->Execute("DELETE FROM GlobalProperties WHERE property=1");      
+      connection_->Execute("INSERT INTO GlobalProperties VALUES (1, '6');");
+      return;  // Success
+    }
+
+#endif
+
+    std::string message = ("Unsupported call to upgrade from version " + boost::lexical_cast<std::string>(version) +
+                           " to version " + boost::lexical_cast<std::string>(targetVersion) + " of the database schema");
+    throw PostgreSQLException(message.c_str());
+  }
+
+
+  void PostgreSQLWrapper::ClearMainDicomTags(int64_t id)
+  {
+    if (clearMainDicomTags_.get() == NULL ||
+        clearDicomIdentifiers_.get() == NULL)
+    {
+      clearMainDicomTags_.reset
+        (new PostgreSQLStatement
+         (*connection_, "DELETE FROM MainDicomTags WHERE id=$1"));
+      clearMainDicomTags_->DeclareInputInteger64(0);
+    
+      clearDicomIdentifiers_.reset
+        (new PostgreSQLStatement
+         (*connection_, "DELETE FROM DicomIdentifiers WHERE id=$1"));
+      clearDicomIdentifiers_->DeclareInputInteger64(0);
+    }
+
+    clearMainDicomTags_->BindInteger64(0, id);
+    clearMainDicomTags_->Run();
+
+    clearDicomIdentifiers_->BindInteger64(0, id);
+    clearDicomIdentifiers_->Run();
+  }
 }
